__author__ = "Vini Salazar"
__license__ = "MIT"
__maintainer__ = "Vini Salazar"
__url__ = "https://github.com/vinisalazar/bioprov"
__version__ = "0.1.1"

"""
Module containing base provenance attributes.

This module extracts system-level information, such as user and environment
settings, and stores them. It is invoked to export provenance objects. 
"""
from os import environ
from bioprov import Project
from bioprov.utils import warnings
from prov.model import ProvDocument, ProvEntity, Namespace, QualifiedName


class BaseProvDocument:
    """
    Class containing base provenance information for a Prov document.

    Adds two default namespaces: 'user', with present user and associated ProvAgent, and 'environment', with
    present environment variables and associated ProvEntity
    """

    def __init__(self, add_namespaces=True):
        """
        Constructor for the base provenance class.
        Creates a prov.model.ProvDocument instance and loads the main attributes.
        :param add_namespaces: Whether to add namespaces when initiating.
        """

        # Initialize document
        self.document = ProvDocument()
        self.env = EnvProv()
        self.user = self.env.user
        self.project = None

        if add_namespaces:
            self._add_namespaces()

    def _add_environ_namespace(self):
        self.document.add_namespace(self.env.env_namespace)
        self.document.entity(
            "env:{}".format(self.env),
            other_attributes=build_prov_attributes(
                self.env.env_dict, self.env.env_namespace
            ),
        )

    def _add_user_namespace(self):
        self.document.add_namespace("user", self.user)
        self.document.agent("user:{}".format(self.user))

    def _add_namespaces(self):
        self._add_environ_namespace()
        self._add_user_namespace()

    def _update_env(self):
        updated_env = self.env.update()
        self.env = updated_env


# class ProjectProv(BaseProvDocument):
#     """
#     Class containing base provenance information for a Project.
#     """
#
#     def __init__(self, project):
#         """
#         Constructs base provenance for a Project.
#         :param project: Project being processed.
#         """
#         # Set Project
#         assert isinstance(project, Project), warnings["incorrect_type"](
#             project, Project
#         )
#         self.project = project
#
#         # Start provenance document
#         self.provdoc = ProvDocument()
#         self.provdoc.add_namespace("user", self.user)
#         self.provdoc.add_namespace("project", self.project.tag.replace(" ", "_"))
#         self.provdoc.add_namespace("samples", str(project).replace(" ", "_"))
#         self.provdoc.add_namespace(
#             "files", "Files_associated_with_project_'{}'".format(project.tag)
#         )
#         self.provdoc.add_namespace(
#             "activities", "Activities_associated_with_project_'{}'".format(project.tag)
#         )
#
#         # Agents
#         self.provdoc.agent("user:{}".format(self.user))
#
#         # Entities
#         self.project.entity = self.provdoc.entity("project:{}".format(project.tag))
#
#         try:
#             self.project_file_entity = self.provdoc.entity(
#                 "files:{}".format(self.project.files["project_csv"].path.name)
#             )
#         except KeyError:
#             raise Exception(
#                 "No 'project_csv' file associated with Project '{}'. Please create a project CSV file.".format(
#                     self.project.tag
#                 )
#             )
#         self.samples_entity = self.provdoc.entity("samples:{}".format(str(project)))
#
#         # Activities
#         self.activities = {
#             "import_Project": self.provdoc.activity(
#                 "activities:{}".format("bioprov.Project")
#             ),
#             "import_Sample": self.provdoc.activity(
#                 "activities:{}".format("bioprov.Sample")
#             ),
#         }
#
#         # Relating project with user, project file, and sample
#         self.provdoc.wasAttributedTo(self.project.entity, "user:{}".format(self.user))
#         self.provdoc.wasGeneratedBy(
#             self.project.entity, self.activities["import_Project"]
#         )
#         self.provdoc.used(self.activities["import_Project"], self.project_file_entity)
#         self.provdoc.used(self.activities["import_Sample"], self.project.entity)
#         self.provdoc.wasGeneratedBy(
#             self.samples_entity, self.activities["import_Sample"]
#         )
#


class EnvProv:
    """
    Class containing provenance information about the current environment
    """

    def __init__(self):

        """
        Class constructor. All attributes are empty and are initialized with self.update()
        """
        self.env_set = None
        self.env_hash = None
        self.env_dict = None
        self.user = None
        self.env_namespace = None
        self.update()

    def __repr__(self):
        return "Environment_hash_'{}'".format(self.env_hash)

    def update(self):
        """
        Checks current environment and updates attributes using the os.environ module.
        :return: Sets attributes to self.
        """
        env_set = frozenset(environ.items())
        env_hash = hash(env_set)
        if env_hash != self.env_hash:
            self.env_set = env_set
            self.env_hash = env_hash
            self.env_dict = dict(self.env_set)
            self.user = self.env_dict["USER"]
            self.env_namespace = Namespace("env", str(self))


def build_prov_attributes(dictionary, namespace):
    """
    :param dictionary: dict with object attributes
    :param namespace: instance of Namespace
    :return: List of tuples (QualifiedName, value)
    """

    # Check arg types
    assert isinstance(namespace, Namespace), warnings["incorrect_type"](
        namespace, Namespace
    )
    assert isinstance(dictionary, dict), warnings["incorrect_type"](dictionary, dict)

    attributes = []
    for k, v in dictionary.items():
        q = QualifiedName(namespace, str(k))
        attributes.append((q, v))

    return attributes
